use std::{borrow::Borrow, fs::File, io::Write, sync::{Arc, Mutex}};

use board::Board;
use ruleset::Ruleset;
use seen::Seen;

use crate::dumb_dealer::Deal;

mod board;
mod ruleset;
mod pico_rng;
mod seen;
mod dumb_dealer;
mod zobrist;


fn main() {
    /*
    let mut rng = PicoRng::srand(0x20000);
    for _ in 0..10 { 
        println!("{}", rng.rnd(0x10000));
    }
    return;
    */

    /*
    let ruleset = Ruleset { 
        n_slots: 11, 
        n_suits: 5, 
        n_cards_per_suit: 10, 
        n_arcana: 25 
    };
    */
    let ruleset = Ruleset { 
        n_slots: 11, 
        n_suits: 4, 
        n_cards_per_suit: 13, 
        n_arcana: 22 
    };
    /* 
    let ruleset = Ruleset { 
        n_slots: 5, 
        n_suits: 1, 
        n_cards_per_suit: 9, 
        n_arcana: 0 
    };
    */
    /*
    let ruleset = Ruleset { 
        n_slots: 7, 
        n_suits: 2, 
        n_cards_per_suit: 9, 
        n_arcana: 8 
    };
    */
    /*
    let ruleset = Ruleset { 
        n_slots: 9, 
        n_suits: 3, 
        n_cards_per_suit: 11, 
        n_arcana: 18 
    };
    */
    let setup = ruleset.compile().expect("compilation should succeed");
    /*
    for _ in 0..10000 {
        Deal::deal(&setup, &mut rand::thread_rng());
    }
    */

    const THREADS: usize = 28;
    let winnable_seeds_file = Arc::new(Mutex::new(File::create("winnable_seeds_multithreaded.txt").expect("should be able to create")));

    let mut threads = Vec::new();
    for i in 0..THREADS {
        let setup2 = setup.clone();
        let wsf = winnable_seeds_file.clone();
        threads.push(std::thread::spawn(move || {
            winnable_seeds_thread(
                &setup2, 
                |seed| {
                    let mut f = wsf.lock().expect("must be able to lock");
                    write!(f, "{}\n", seed).expect("write should succeed");
                    f.flush().expect("flush should succeed");
                },
                i as u32, 
                THREADS as u32
            );
        }));
    }

    for i in threads {
        let _ = i.join();
    }

}

fn winnable_seeds_thread(setup: &ruleset::Setup, mut cb: impl FnMut(u32), start: u32, step: u32) {
    let mut winnable = 0;
    let mut total = 0;
    
    for seed in (start..0xffffffff).step_by(step as usize) {
        let mut board = Board::new(&setup);
        board.deal(Deal::deal(&setup, seed));
        board.display();

        if is_winnable(board) {
            winnable += 1;
            cb(seed);
        }
        total += 1;
        println!("winnable: {}/{} ({}%)", winnable, total, (100.0 * winnable as f32)/(total as f32));
    }

    // println!("Legal moves: {:#?}", board.legal_moves());

}

fn is_winnable(mut board: Board<'_>) -> bool {
    return 
        explore(20, &mut board)  ||
        explore(200, &mut board)
    ;
}

fn explore(max_depth: usize, board: &mut Board<'_>) -> bool {
    let mut seen = Seen::new();
    explore2(max_depth, 0, board, &mut seen)
}

fn explore2(max_depth: usize, depth: usize, board: &mut Board<'_>, seen: &mut Seen) -> bool {
    if depth > max_depth {
        return false;
    }

    if seen.contains(board.zobrist_key()) {
        return false
    }
    seen.add(board.zobrist_key());

    if board.is_won() {
        board.display();
        return true
    }

    for m in board.legal_moves() {
        let hash_1 = board.zobrist_key();
        board.perform(m);
        // println!("try: {:X?} {:?}", board.zobrist_key(), m);
        if explore2(max_depth, depth + 1, board, seen) {
            return true;
        }
        // println!("undo: {:X?} {:?}", board.zobrist_key(), m);
        board.undo();
        let hash_2 = board.zobrist_key();
        assert_eq!(hash_1, hash_2)
    }
    return false;
}